# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import logging
import time
import threading


### For svg parsing

# requires svg.path, install it like this: pip3 install svg.path
# converts a list of path elements of a SVG file to simple line drawing commands
try:
    from svg.path import parse_path
    from svg.path.path import Line
    from xml.dom import minidom
except ImportError:
    parse_path = None

### End of import for svg parsing

try:
    # pylint: disable=import-error
    import win32print
except ImportError:
    logging.getLogger('WinDriverDetector').warning('No win32print installed')
    win32print = None

from base_sender import BaseSender
import platforms


class Sender(BaseSender):

    FILENAME = '1.ps'
    PASS_TO_PRINTER_CONSTANT = "XPS_PASS"
    OLD_PASS_TO_PRINTER_CONSTANT = "RAW"
    FLUSH_SLEEP = 3000 #Number of milliseconds to suspend printer
    WIN_CALLS_LEVEL = 2

    def __init__(self, parent, usb_info={}, profile={}):
        if not win32print:
            raise RuntimeError('No backend module to use this printer')
        self.logger = logging.getLogger(self.__class__.__name__)
        self.printer_handle = None
        self.stop_flag = False
        self.status_int = 0x00000008
        self.job_id = None
        self.profile = profile
        self.printer_name = usb_info.get('MSNAME')
        if not win32print:
            raise RuntimeError('Unable to get a driver for a printer')
        if not self.printer_name:
            raise RuntimeError('No printer name for MS Windows printer')
        if self.connect():
            self.print_handle_lock = threading.RLock()
            self.state_requesting_thread = threading.Thread(target=self.state_requesting)
            self.state_requesting_thread.start()
        else:
            raise RuntimeError('Unable to connect to MS Windows printer ' + self.printer_name)

    def get_raw_type_name(self):
        drivers = win32print.EnumPrinterDrivers(None, None, self.WIN_CALLS_LEVEL)
        printer_info = win32print.GetPrinter(self.printer_handle, self.WIN_CALLS_LEVEL)
        for driver in drivers:
            if driver["Name"] == printer_info["pDriverName"]:
                if driver["Version"] == 4:
                    return self.PASS_TO_PRINTER_CONSTANT
                else:
                    return self.OLD_PASS_TO_PRINTER_CONSTANT
        raise RuntimeError('Unable to get a driver for a printer')

    def connect(self):
        try:
            self.printer_handle = win32print.OpenPrinter(self.printer_name)
            self.raw_type_name = self.get_raw_type_name()
            return True
        except:
            self.logger.info('MS Windows error on opening printer: ' + str(self.printer_handle))
            return False

    def parse_printer_info_dict(self, printer_info_dict):
        #TODO
        pass

    def state_requesting(self):
        got_printer_handle = bool(self.printer_handle)
        while not self.stop_flag:
            if not self.printer_handle:
                if got_printer_handle:
                    self.logger.warning('')
                    got_printer_handle = False
            else:
                if not got_printer_handle:
                    got_printer_handle = True
                printer_info_dict = win32print.GetPrinter(self.printer_handle, self.WIN_CALLS_LEVEL)
                self.parse_printer_info_dict(printer_info_dict)

    def load_gcodes(self, f_or_gcodes):
        #TODO
        return True

    def close(self):
        if self.printer_handle:
            win32print.ClosePrinter(self.printer_handle)

    def flush_printer(self):
        win32print.FlushPrinter(self.printer_handle, '', self.FLUSH_SLEEP)

    #NOT USED
    # def set_properties(self):
    #     # Get the default properties for the printer

    #     # Constants from wingdi.h
    #     DM_OUT_BUFFER = 0x02
    #     DM_IN_BUFFER = 0x08
    #     DM_IN_PROMPT = 0x04
    #     DM_DEFAULT_SOURCE = 0x200

    #     properties = win32print.GetPrinter(self.printer_handle, self.WIN_CONST_FOR_API_CALLS)
    #     self.logger.debug('Properties', properties)
    #     devmode = properties['pDevMode']
    #     #Print the default paper source (prints '7' for 'Automatically select')
    #     self.logger.debug(devmode.DefaultSource)

    #     # Change the default paper source to '4' for 'Manual feed'
    #     devmode.DefaultSource = 4
    #     devmode.Fields = devmode.Fields | DM_DEFAULT_SOURCE

    #     # Write these changes back to the printer
    #     win32print.DocumentProperties(None, handle, device_name, devmode, devmode, DM_IN_BUFFER | DM_OUT_BUFFER)

    def print(self, f, filetype_name):
        docname = self.filename
        if not docname:
            docname = '3DPrinterOSJob'
        job_id = win32print.StartDocPrinter(self.printer_handle, 1, (docname, None, self.raw_type_name))
        self.logger.debug('MSWin job id:' + str(job_id))
        if job_id:
            self.job_id = job_id
            win32print.StartPagePrinter(self.printer_handle)
            buffer = f.read()
            self.logger.debug('Print document of len :' + str(len(buffer)))
            win32print.EndPagePrinter(self.printer_handle)
            bytes_written = win32print.WritePrinter(self.printer_handle, buffer)
            print('Written:' + str(bytes_written))
            win32print.EndDocPrinter(self.printer_handle)
            win32print.ClosePrinter(self.printer_handle)

    def parse_svg_to_path_strings(self, svg_file_path):
        points = []
        if parse_path:
            try:
                doc = minidom.parse(svg_file_path)
                path_strings = [path.getAttribute('d') for path in doc.getElementsByTagName('path')]
                doc.unlink()
            except Exception as e:
                self.logger.warning('Error loading svg file: ' + str(svg_file_path))
            else:
                for path_string in path_strings:
                    path = parse_path(path_string)
                    for e in path:
                        if isinstance(e, Line):
                            x0 = e.start.real
                            y0 = e.start.imag
                            x1 = e.end.real
                            y1 = e.end.imag
                            points.append( (x0, y0), (x1, y1) )
        return points

    """
    Error

        define PRINTER_STATUS_PAUSED            0x00000001
        define PRINTER_STATUS_ERROR             0x00000002
        define PRINTER_STATUS_PENDING_DELETION  0x00000004
        define PRINTER_STATUS_PAPER_JAM         0x00000008
        define PRINTER_STATUS_PAPER_OUT         0x00000010
        define PRINTER_STATUS_MANUAL_FEED       0x00000020
        define PRINTER_STATUS_PAPER_PROBLEM     0x00000040
        define PRINTER_STATUS_OFFLINE           0x00000080
        define PRINTER_STATUS_IO_ACTIVE         0x00000100
        define PRINTER_STATUS_BUSY              0x00000200
        define PRINTER_STATUS_PRINTING          0x00000400
        define PRINTER_STATUS_OUTPUT_BIN_FULL   0x00000800
        define PRINTER_STATUS_NOT_AVAILABLE     0x00001000
        define PRINTER_STATUS_WAITING           0x00002000
        define PRINTER_STATUS_PROCESSING        0x00004000
        define PRINTER_STATUS_INITIALIZING      0x00008000
        define PRINTER_STATUS_WARMING_UP        0x00010000
        define PRINTER_STATUS_TONER_LOW         0x00020000
        define PRINTER_STATUS_NO_TONER          0x00040000
        define PRINTER_STATUS_PAGE_PUNT         0x00080000
        define PRINTER_STATUS_USER_INTERVENTION 0x00100000
        define PRINTER_STATUS_OUT_OF_MEMORY     0x00200000
        define PRINTER_STATUS_DOOR_OPEN         0x00400000
        define PRINTER_STATUS_SERVER_UNKNOWN    0x00800000
        define PRINTER_STATUS_POWER_SAVE        0x01000000
        define PRINTER_STATUS_SERVER_OFFLINE    0x02000000
        define PRINTER_STATUS_DRIVER_UPDATE_NEEDED    0x04000000


    Control

        define PRINTER_CONTROL_PAUSE            1
        define PRINTER_CONTROL_RESUME           2
        define PRINTER_CONTROL_PURGE            3
        define PRINTER_CONTROL_SET_STATUS       4


    Status

        The printer status. This member can be any reasonable combination of the following values.
        Value 	Meaning
        PRINTER_STATUS_BUSY 	The printer is busy.
        PRINTER_STATUS_DOOR_OPEN 	The printer door is open.
        PRINTER_STATUS_ERROR 	The printer is in an error state.
        PRINTER_STATUS_INITIALIZING 	The printer is initializing.
        PRINTER_STATUS_IO_ACTIVE 	The printer is in an active input/output state
        PRINTER_STATUS_MANUAL_FEED 	The printer is in a manual feed state.
        PRINTER_STATUS_NO_TONER 	The printer is out of toner.
        PRINTER_STATUS_NOT_AVAILABLE 	The printer is not available for printing.
        PRINTER_STATUS_OFFLINE 	The printer is offline.
        PRINTER_STATUS_OUT_OF_MEMORY 	The printer has run out of memory.
        PRINTER_STATUS_OUTPUT_BIN_FULL 	The printer's output bin is full.
        PRINTER_STATUS_PAGE_PUNT 	The printer cannot print the current page.
        PRINTER_STATUS_PAPER_JAM 	Paper is jammed in the printer
        PRINTER_STATUS_PAPER_OUT 	The printer is out of paper.
        PRINTER_STATUS_PAPER_PROBLEM 	The printer has a paper problem.
        PRINTER_STATUS_PAUSED 	The printer is paused.
        PRINTER_STATUS_PENDING_DELETION 	The printer is being deleted.
        PRINTER_STATUS_POWER_SAVE 	The printer is in power save mode.
        PRINTER_STATUS_PRINTING 	The printer is printing.
        PRINTER_STATUS_PROCESSING 	The printer is processing a print job.
        PRINTER_STATUS_SERVER_UNKNOWN 	The printer status is unknown.
        PRINTER_STATUS_TONER_LOW 	The printer is low on toner.
        PRINTER_STATUS_USER_INTERVENTION 	The printer has an error that requires the user to do something.
        PRINTER_STATUS_WAITING 	The printer is waiting.
        PRINTER_STATUS_WARMING_UP 	The printer is warming up.


    hPrinter : PyPrinterHANDLE

        handle to printer object as returned by win32print::OpenPrinter

    Level=2 : int

        Level of data returned (1,2,3,4,5,7,8,9)

        Comments
        Original implementation used level 2 only and returned a tuple Pass single arg as indicator to use old behaviour for backward compatibility

        Return Value
        Returns a dictionary containing PRINTER_INFO_* data for level, or returns a tuple of PRINTER_INFO_2 data if no level is passed in.
        """

