# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import time
import json
import select
import socket
import logging
import threading
import os.path

import config
import usb_detect
import paths
import http_rrfw_detect
import http_reperier_detect
import zeroconf_detect
import upnp_detect
import broadcast_printer_detector


# TODO refactor 
# 1) use multithreading or asyncio to scan simultaneously
# 2) decouple starting, running and processing from network detector and other detectors
# 3) refactor to simplify network detector logic and minimize this class

SCANNERS = {'broadcast': broadcast_printer_detector.BroadcastPrinterDetector,\
            'rrfw' : http_rrfw_detect.HTTPRRFWDetector,\
            'repetier' : http_reperier_detect.HTTPRepetierDetector,\
            'zeroconf' : zeroconf_detect.ZeroconfDetector,\
            'upnp' : upnp_detect.UPNPDetector}
class NetworkDetector:

    STORAGE_FILE_PATH = os.path.join(paths.CURRENT_SETTINGS_FOLDER, 'detected_printers.json')
    DEFAULT_SCANNER_CLASS = broadcast_printer_detector.BroadcastPrinterDetector

    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger(self.__class__.__name__)
        self.detected_printers = []
        self.printers_add_by_ip = []
        self.load_printers_list()
        self.network_detector_run_flag = False

    def refresh_on_next_loop(self):
        self.network_detector_run_flag = True

    def parent_stop(self):
        return self.parent and self.parent.stop_flag

    def get_all_network_profiles(self):
        return [x for x in config.get_profiles() if x.get('network_detect')]

    def get_unique_network_profiles(self):
        profiles = []
        network_detects = []
        for profile in self.get_all_network_profiles():
            if profile['network_detect'] not in network_detects:
                profiles.append(profile)
                network_detects.append(profile['network_detect'])
        return profiles

    def get_printers_list(self):
        if self.network_detector_run_flag:
            self.logger.info('Scanning for network printers...')
            self.detected_printers = []
            for profile in self.get_unique_network_profiles():
                if self.parent_stop():
                    return []
                scanner_name = profile.get('network_detect', {}).get('scanner')
                scanner_class = SCANNERS.get(scanner_name)
                if not scanner_class:
                    self.logger.error(f'Unknown scanner name: {scanner_name}')
                    continue
                scanner = scanner_class(self.parent, profile)
                scanner.detect()
                self.logger.debug('Scanner:\t%s\t\%s' % (scanner, scanner.discovered_printers))
                self.detected_printers.extend(scanner.discovered_printers)
            self.logger.info('Discovered printers:\n%s' % self.detected_printers)
            self.save_printers_list()
            self.network_detector_run_flag = False
        return self.detected_printers + self.printers_add_by_ip

    def printers_list_with_profiles(self):
        printers_list = []
        for info in self.get_printers_list():
            for pi in getattr(self.parent, "printer_interfaces", []):
                if pi.usb_info == info:
                    printers_list.append({"usb_info": info, "profile": pi.printer_profile})
                    break
            else:
                printers_list.append({"usb_info": info, "profile": {}})
        return printers_list

    def add_network_printer(self, printer_ip, printer_type, printer_port=None, run_detector=True):
        for profile in self.get_all_network_profiles():
            if printer_type == profile.get("name"):
                scanner_name = profile.get('network_detect', {}).get('scanner')
                if run_detector and scanner_name: #disable scan for birdwing printers
                    if scanner_name in SCANNERS:
                        scanner_class = SCANNERS[scanner_name]
                        scanner = scanner_class(self.parent, profile)
                        scanner.detect(already_know_ip=printer_ip, non_default_port=printer_port)
                        for printer in scanner.discovered_printers:
                            if printer not in self.printers_add_by_ip and not printer in self.detected_printers:
                                self.printers_add_by_ip.append(printer)
                        self.save_printers_list()
                        break
                    else:
                        self.logger.warning(f'Unknown scanner: {scanner_name}')
                else:
                    printer = {
                        'IP': printer_ip,
                        'SNR': printer_ip,
                        'VID': profile["vids_pids"][0][0],
                        'PID': profile["vids_pids"][0][1]
                    }
                    if printer not in self.printers_add_by_ip:
                        self.printers_add_by_ip.append(printer)
                        self.save_printers_list()
                    break

    def forget(self, usb_info):
        for group in (self.detected_printers, self.printers_add_by_ip):
            for printer in group:
                if printer == usb_info:
                    group.remove(printer)
        self.save_printers_list()

    def save_printers_list(self):
        file_path = self.STORAGE_FILE_PATH
        try:
            with open(file_path, 'w') as f:
                data = {'detected': self.detected_printers, 'added_by_ip': self.printers_add_by_ip}
                json_config = json.dumps(data, indent = 4, separators = (',', ': '))
                f.write(json_config)
        except Exception as e:
            self.logger.error("Error writing data to %s: %s" % (file_path, str(e)))

    def load_printers_list(self):
        file_path = self.STORAGE_FILE_PATH
        try:
            with open(file_path, 'r') as f:
                data = f.read()
                self.logger.info("Stored printers list: %s" % data)
                printers_list = json.loads(data)
                self.detected_printers = printers_list.get('detected', [])
                self.printers_add_by_ip = printers_list.get('added_by_ip', [])
        except FileNotFoundError:
            pass
        except Exception as e:
            self.logger.error("Error reading data from %s: %s" % (file_path, str(e)))

if __name__ == '__main__':
    from pprint import pprint

    logging.basicConfig(level=logging.DEBUG)
    nd = NetworkDetector()
    nd.refresh_on_next_loop()
    print("Detected network printers:")
    nd.get_printers_list()
